 /*
   ESRI Shapefile extension for HHVM.
   Copyright (c) 2013-2020 SIA Intelligent Systems.
   
   Based on PECL shape extension with the following copyright:

   +----------------------------------------------------------------------+
   | PHP Version 5                                                        |
   +----------------------------------------------------------------------+
   | Copyright (c) 1997-2005 The PHP Group                                |
   +----------------------------------------------------------------------+
   | This source file is subject to version 3.0 of the PHP license,       |
   | that is bundled with this package in the file LICENSE, and is        |
   | available through the world-wide-web at the following url:           |
   | http://www.php.net/license/3_0.txt.                                  |
   | If you did not receive a copy of the PHP license and are unable to   |
   | obtain it through the world-wide-web, please send a note to          |
   | license@php.net so we can mail you a copy immediately.               |
   +----------------------------------------------------------------------+
   | Author: Andrey Hristov                                               |
   +----------------------------------------------------------------------+

   Extended and ported to HHVM by Kristaps Kaupe.
*/

#include "hphp/runtime/version.h"
#include "hphp/runtime/ext/extension.h"
#include "hphp/runtime/base/array-iterator.h"
#include "hphp/runtime/base/resource-data.h"
#include <cinttypes>
#include <libshp/shapefil.h>

namespace HPHP {

// HPHP::Array::add() was removed in HHVM 3.29, should use set() instead
#if defined HHVM_VERSION_BRANCH && HHVM_VERSION_BRANCH >= 0x031D00
  #define ARRAY_ADD(arr,key,value) arr.set(key, Variant(value))
#else
  #define ARRAY_ADD(arr,key,value) arr.add(key, Variant(value))
#endif

const StaticString s_SHPT_NULL("SHPT_NULL");
const int64_t k_SHPT_NULL = SHPT_NULL;
const StaticString s_SHPT_POINT("SHPT_POINT");
const int64_t k_SHPT_POINT = SHPT_POINT;
const StaticString s_SHPT_ARC("SHPT_ARC");
const int64_t k_SHPT_ARC = SHPT_ARC;
const StaticString s_SHPT_POLYGON("SHPT_POLYGON");
const int64_t k_SHPT_POLYGON = SHPT_POLYGON;
const StaticString s_SHPT_MULTIPOINT("SHPT_MULTIPOINT");
const int64_t k_SHPT_MULTIPOINT = SHPT_MULTIPOINT;
const StaticString s_SHPT_POINTZ("SHPT_POINTZ");
const int64_t k_SHPT_POINTZ = SHPT_POINTZ;
const StaticString s_SHPT_ARCZ("SHPT_ARCZ");
const int64_t k_SHPT_ARCZ = SHPT_ARCZ;
const StaticString s_SHPT_POLYGONZ("SHPT_POLYGONZ");
const int64_t k_SHPT_POLYGONZ = SHPT_POLYGONZ;
const StaticString s_SHPT_MULTIPOINTZ("SHPT_MULTIPOINTZ");
const int64_t k_SHPT_MULTIPOINTZ = SHPT_MULTIPOINTZ;
const StaticString s_SHPT_POINTM("SHPT_POINTM");
const int64_t k_SHPT_POINTM = SHPT_POINTM;
const StaticString s_SHPT_ARCM("SHPT_ARCM");
const int64_t k_SHPT_ARCM = SHPT_ARCM;
const StaticString s_SHPT_POLYGONM("SHPT_POLYGONM");
const int64_t k_SHPT_POLYGONM = SHPT_POLYGONM;
const StaticString s_SHPT_MULTIPOINTM("SHPT_MULTIPOINTM");
const int64_t k_SHPT_MULTIPOINTM = SHPT_MULTIPOINTM;
const StaticString s_SHPT_MULTIPATCH("SHPT_MULTIPATCH");
const int64_t k_SHPT_MULTIPATCH = SHPT_MULTIPATCH;
const StaticString s_SHPP_TRISTRIP("SHPP_TRISTRIP");
const int64_t k_SHPP_TRISTRIP = SHPP_TRISTRIP;
const StaticString s_SHPP_TRIFAN("SHPP_TRIFAN");
const int64_t k_SHPP_TRIFAN = SHPP_TRIFAN;
const StaticString s_SHPP_OUTERRING("SHPP_OUTERRING");
const int64_t k_SHPP_OUTERRING = SHPP_OUTERRING;
const StaticString s_SHPP_INNERRING("SHPP_INNERRING");
const int64_t k_SHPP_INNERRING = SHPP_INNERRING;
const StaticString s_SHPP_FIRSTRING("SHPP_FIRSTRING");
const int64_t k_SHPP_FIRSTRING = SHPP_FIRSTRING;
const StaticString s_SHPP_RING("SHPP_RING");
const int64_t k_SHPP_RING = SHPP_RING;

class SHPHandleRes : public SweepableResourceData {
public:
  explicit SHPHandleRes(SHPHandle handle) : shp_handle(handle) { };
  SHPHandle operator* () const { return shp_handle; };
  virtual void sweep() { delete this; }

  static StaticString s_class_name;
  virtual const String& o_getClassNameHook() const { return s_class_name; }

  SHPHandle shp_handle;
};

StaticString SHPHandleRes::s_class_name("SHPHandle");

class SHPObjectRes : public SweepableResourceData {
public:
  explicit SHPObjectRes(SHPObject* object) : shp_object(object) { };
  SHPObject* operator* () const { return shp_object; };
  virtual void sweep() { delete this; }

  static StaticString s_class_name;
  virtual const String& o_getClassNameHook() const { return s_class_name; }

  SHPObject* shp_object;
};

StaticString SHPObjectRes::s_class_name("SHPObject");

#if (HHVM_VERSION_BRANCH >= ((3 << 16) | (10 << 8)))

  #define CHECK_HANDLE(handle, type, shp, ret) \
    req::ptr<type ## Res> r_ ## shp = cast<type ## Res>(handle); \
    if (! r_ ## shp) { \
      raise_warning("Got NULL for resource"); \
      return (ret); \
    } else { \
      shp = **r_ ## shp.get(); \
    }

  #define NEWRES(x) req::make<x>
  #define DETACH .detach()

#else

  #define CHECK_HANDLE(handle, type, shp, ret) \
    type ## Res * r_ ## shp = handle.getTyped<type ## Res>(true, true); \
    if (r_ ## shp == nullptr) { \
      raise_warning("Got NULL for resource"); \
      return (ret); \
    } else { \
      shp = **(r_ ## shp); \
    }

  #ifdef NEWOBJ
    #define NEWRES(x) NEWOBJ(x)
  #else
    #define NEWRES(x) newres<x>
  #endif
  #define DETACH

#endif

static Variant HHVM_FUNCTION(shp_open, const String& filename, const String& access) {
  SHPHandle shph = SHPOpen(filename.c_str(), access.c_str());

  if (!shph) {
    return Variant(Variant::NullInit{});
  }

  SHPHandleRes* shph_res = NEWRES(SHPHandleRes)(shph) DETACH;
  return Resource(shph_res);
}

static Variant HHVM_FUNCTION(shp_create, const String& filename, int64_t shape_type) {
  switch (shape_type) {
  case SHPT_POINT:
  case SHPT_ARC:
  case SHPT_POLYGON:
  case SHPT_MULTIPOINT:
    break;
  default:
    raise_warning("Unsupported type %d", (int)shape_type);
    break;
  }
  SHPHandle shph = SHPCreate(filename.c_str(), shape_type);

  if (!shph) {
    return Variant(Variant::NullInit{});
  }

  SHPHandleRes* shph_res = NEWRES(SHPHandleRes)(shph) DETACH;
  return Resource(shph_res);
}

static bool HHVM_FUNCTION(shp_close, const Resource& shp) {
  // We can have invalid pointer here. Better leak memory than crash.
  //SHPHandle shph;
  //CHECK_HANDLE(shp, SHPHandle, shph, false);
  //SHPClose(shph);
  return true;
}

static Variant HHVM_FUNCTION(shp_read_object, const Resource& shp, int64_t ord) {
  SHPHandle shph;
  CHECK_HANDLE(shp, SHPHandle, shph, false);

  SHPObject* shp_obj = SHPReadObject(shph, ord);
  if (!shp_obj) {
    raise_warning("Reading of object %d failed", (int)ord);
    return Variant(Variant::NullInit{});
  }

  SHPObjectRes* shp_obj_res = NEWRES(SHPObjectRes)(shp_obj) DETACH;
  return Resource(shp_obj_res);
}

static bool HHVM_FUNCTION(shp_destroy_object, const Resource& shp_object) {
  // We can have invalid pointer here. Better leak memory than crash.
  //SHPObject* shp_obj;
  //CHECK_HANDLE(shp_object, SHPObject, shp_obj, false);
  //SHPDestroyObject(shp_obj);
  return true;
}

static bool HHVM_FUNCTION(shp_rewind_object, const Resource& shp_handle, const Resource& shp_object) {
  SHPHandle shph;
  CHECK_HANDLE(shp_handle, SHPHandle, shph, false);
  SHPObject* shp_obj;
  CHECK_HANDLE(shp_object, SHPObject, shp_obj, false);
  return SHPRewindObject(shph, shp_obj);
}

static Variant HHVM_FUNCTION(shp_write_object, const Resource& shp_handle, int64_t entity_num, const Resource& shp_object) {
  SHPHandle shph;
  CHECK_HANDLE(shp_handle, SHPHandle, shph, false);
  SHPObject* shp_obj;
  CHECK_HANDLE(shp_object, SHPObject, shp_obj, false);
  
  // Do the same checks as asserts in SHPWriteObject()
  if (shp_obj->nSHPType != shph->nShapeType && shp_obj->nSHPType != SHPT_NULL) {
    raise_warning(
      "Shape object does not match the type of the file it is being written to"
    );
    return Variant(false);
  }
  if (entity_num != -1 && entity_num >= shph->nRecords) {
    raise_warning(
      "shp_write_object() expects entity_num -1 for appends, forcing"
    );
    entity_num = -1;
  }
  
  return Variant(SHPWriteObject(shph, entity_num, shp_obj));
}

static Variant HHVM_FUNCTION(shp_get_info, const Resource& shp_handle) {
  SHPHandle shph;
  CHECK_HANDLE(shp_handle, SHPHandle, shph, false);

  int ent_n, shp_type;
  double min_b[4], max_b[4];
  SHPGetInfo(shph, &ent_n, &shp_type, min_b, max_b);
  Array min_bound = Array::Create();
  Array max_bound = Array::Create();
  for (int i = 0; i < 4; i++) {
    ARRAY_ADD(min_bound, i, min_b[i]);
    ARRAY_ADD(min_bound, i, max_b[i]);
  }
  Array return_value = Array::Create();
  ARRAY_ADD(return_value, String("pnEntities"), ent_n);
  ARRAY_ADD(return_value, String("pnShapetype"), shp_type);
  ARRAY_ADD(return_value, String("padfMinBound"), min_bound);
  ARRAY_ADD(return_value, String("padfMaxBound"), max_bound);
  return Variant(return_value);
}

static bool HHVM_FUNCTION(shp_compute_extents, const Resource& shp_object) {
  SHPObject* shp_obj;
  CHECK_HANDLE(shp_object, SHPObject, shp_obj, false);

  SHPComputeExtents(shp_obj);
  return true;
}

static Variant HHVM_FUNCTION(shp_create_object_impl, const Array& args) {
  if (args.size() < 10) {
    return Variant(false);
  }
  int shp_type = args[0].toInt32(10);
  int shp_id = args[1].toInt32(10);
  int parts_num = args[2].toInt32(10);
  Array panPartStart = args[3].toArray();
  Array panPartType = args[4].toArray();
  int vertices_num = args[5].toInt32(10);
  Array padfX = args[6].toArray();
  Array padfY = args[7].toArray();
  Array padfZ = args[8].toArray();
  Array padfM = args[9].toArray();

  if (padfX.size() != padfY.size()) {
    raise_warning("padfX and padfY have different number of elements");
    return Variant(false);
  }
  if (padfX.size() < vertices_num) {
    raise_warning("less elements in padfX than specified in vertices_num");
    return Variant(false);
  }

  int c_panPartStart[panPartStart.size()];
  {
    size_t i = 0;
    for (ArrayIter arr_it(panPartStart); arr_it; ++arr_it, i++) {
      c_panPartStart[i] = arr_it.second().toInt32(10);
    }
  }

  int c_panPartType[panPartType.size()];
  {
    size_t i = 0;
    for (ArrayIter arr_it(panPartType); arr_it; ++arr_it, i++) {
      c_panPartType[i] = arr_it.second().toInt32(10);
    }
  }

  assert(padfX.size() == padfY.size());
  ssize_t padfZ_size = padfZ.size();
  if (padfZ_size == 0) { padfZ_size = SSIZE_MAX; }
  int real_vert_num = std::min(padfX.size(), padfZ_size);
  if (real_vert_num != vertices_num) {
    raise_notice("The real number of used vertices %d, passed %d", real_vert_num, vertices_num);
  }

  double c_padfX[padfX.size()];
  double c_padfY[padfY.size()];
  double c_padfZ[padfZ.size()];
  double c_padfM[padfM.size()];
  if (real_vert_num > 0) {
    {
      size_t i = 0;
      for (ArrayIter arr_it(padfX); arr_it; ++arr_it, i++) {
        c_padfX[i] = arr_it.second().toDouble();
      }
    }
    {
      size_t i = 0;
      for (ArrayIter arr_it(padfY); arr_it; ++arr_it, i++) {
        c_padfY[i] = arr_it.second().toDouble();
      }
    }
    {
      size_t i = 0;
      for (ArrayIter arr_it(padfZ); arr_it; ++arr_it, i++) {
        c_padfZ[i] = arr_it.second().toDouble();
      }
    }
    {
      size_t i = 0;
      for (ArrayIter arr_it(padfM); arr_it; ++arr_it, i++) {
        c_padfM[i] = arr_it.second().toDouble();
      }
    }
  }

  SHPObject* shp_obj = SHPCreateObject(shp_type, shp_id, parts_num, c_panPartStart, c_panPartType,
    real_vert_num, c_padfX, c_padfY, c_padfZ, c_padfM);
  if (shp_obj == nullptr) {
    raise_warning("SHPCreateObject returned NULL");
    return Variant(false);
  }

  SHPObjectRes* shp_obj_res = NEWRES(SHPObjectRes)(shp_obj) DETACH;
  return Resource(shp_obj_res);
}

static Variant HHVM_FUNCTION(shp_create_simple_object,
  int64_t shp_type, int64_t vertices_num,
  const Array& padfX, const Array& padfY, const Array& padfZ
) {
  if (padfX.size() != padfY.size()) {
    raise_warning("padfX and padfY have different number of elements");
    return Variant(false);
  }
  if (padfX.size() < vertices_num) {
    raise_warning("less elements in padfX than specified in vertices_num");
    return Variant(false);
  }

  assert(padfX.size() == padfY.size());
  ssize_t padfZ_size = padfZ.size();
  if (padfZ_size == 0) { padfZ_size = SSIZE_MAX; }
  int real_vert_num = std::min(padfX.size(), padfZ_size);
  if (real_vert_num != vertices_num) {
    raise_notice("The real number of used vertices %d, passed %d", real_vert_num, (int)vertices_num);
  }

  double c_padfX[padfX.size()];
  double c_padfY[padfY.size()];
  double c_padfZ[padfZ.size()];
  if (real_vert_num > 0) {
    {
      size_t i = 0;
      for (ArrayIter arr_it(padfX); arr_it; ++arr_it, i++) {
        c_padfX[i] = arr_it.second().toDouble();
      }
    }
    {
      size_t i = 0;
      for (ArrayIter arr_it(padfY); arr_it; ++arr_it, i++) {
        c_padfY[i] = arr_it.second().toDouble();
      }
    }
    {
      size_t i = 0;
      for (ArrayIter arr_it(padfZ); arr_it; ++arr_it, i++) {
        c_padfZ[i] = arr_it.second().toDouble();
      }
    }
  }

  SHPObject* shp_obj = SHPCreateSimpleObject(shp_type, real_vert_num, c_padfX, c_padfY, c_padfZ);
  if (shp_obj == nullptr) {
    raise_warning("SHPCreateSimpleObject returned NULL");
    return Variant(false);
  }

  SHPObjectRes* shp_obj_res = NEWRES(SHPObjectRes)(shp_obj) DETACH;
  return Resource(shp_obj_res);
}

static Variant HHVM_FUNCTION(shp_get_array_from_object, const Resource& shp_object) {
  SHPObject* shp_obj;
  CHECK_HANDLE(shp_object, SHPObject, shp_obj, false);

  Array return_value = Array::Create();
  ARRAY_ADD(return_value, String("nVertices"), shp_obj->nVertices);
  Array pointX_arr = Array::Create();
  Array pointY_arr = Array::Create();
  for (int i = 0; i < shp_obj->nVertices; i++) {
    ARRAY_ADD(pointX_arr, i, shp_obj->padfX[i]);
    ARRAY_ADD(pointY_arr, i, shp_obj->padfY[i]);
  }
  ARRAY_ADD(return_value, String("padfX"), pointX_arr);
  ARRAY_ADD(return_value, String("padfY"), pointY_arr);
  return Variant(return_value);
}

//////////////////////////////////////////////////////////////////////////////

class shpExtension: public Extension {
public:
  // 0.9.2-dev is PECL extension version we ported.
  shpExtension(): Extension("shp", "0.9.2-dev") { /* nothing */ }
  virtual void moduleInit() override {
    Native::registerConstant<KindOfInt64>(s_SHPT_NULL.get(), k_SHPT_NULL);
    Native::registerConstant<KindOfInt64>(s_SHPT_POINT.get(), k_SHPT_POINT);
    Native::registerConstant<KindOfInt64>(s_SHPT_ARC.get(), k_SHPT_ARC);
    Native::registerConstant<KindOfInt64>(s_SHPT_POLYGON.get(), k_SHPT_POLYGON);
    Native::registerConstant<KindOfInt64>(s_SHPT_MULTIPOINT.get(), k_SHPT_MULTIPOINT);
    Native::registerConstant<KindOfInt64>(s_SHPT_POINTZ.get(), k_SHPT_POINTZ);
    Native::registerConstant<KindOfInt64>(s_SHPT_ARCZ.get(), k_SHPT_ARCZ);
    Native::registerConstant<KindOfInt64>(s_SHPT_POLYGONZ.get(), k_SHPT_POLYGONZ);
    Native::registerConstant<KindOfInt64>(s_SHPT_MULTIPOINTZ.get(), k_SHPT_MULTIPOINTZ);
    Native::registerConstant<KindOfInt64>(s_SHPT_POINTM.get(), k_SHPT_POINTM);
    Native::registerConstant<KindOfInt64>(s_SHPT_ARCM.get(), k_SHPT_ARCM);
    Native::registerConstant<KindOfInt64>(s_SHPT_POLYGONM.get(), k_SHPT_POLYGONM);
    Native::registerConstant<KindOfInt64>(s_SHPT_MULTIPOINTM.get(), k_SHPT_MULTIPOINTM);
    Native::registerConstant<KindOfInt64>(s_SHPT_MULTIPATCH.get(), k_SHPT_MULTIPATCH);
    Native::registerConstant<KindOfInt64>(s_SHPP_TRISTRIP.get(), k_SHPP_TRISTRIP);
    Native::registerConstant<KindOfInt64>(s_SHPP_TRIFAN.get(), k_SHPP_TRIFAN);
    Native::registerConstant<KindOfInt64>(s_SHPP_OUTERRING.get(), k_SHPP_OUTERRING);
    Native::registerConstant<KindOfInt64>(s_SHPP_INNERRING.get(), k_SHPP_INNERRING);
    Native::registerConstant<KindOfInt64>(s_SHPP_FIRSTRING.get(), k_SHPP_FIRSTRING);
    Native::registerConstant<KindOfInt64>(s_SHPP_RING.get(), k_SHPP_RING);
    HHVM_FE(shp_open);
    HHVM_FE(shp_create);
    HHVM_FE(shp_close);
    HHVM_FE(shp_read_object);
    HHVM_FE(shp_destroy_object);
    HHVM_FE(shp_rewind_object);
    HHVM_FE(shp_write_object);
    HHVM_FE(shp_get_info);
    HHVM_FE(shp_compute_extents);
    HHVM_FE(shp_create_object_impl);
    HHVM_FE(shp_create_simple_object);
    HHVM_FE(shp_get_array_from_object);
    loadSystemlib();
  }
} s_shp_extension;

HHVM_GET_MODULE(shp);

}

